"""Deserializer for an enum."""

import enum
from functools import partial
from typing import Any, List, Sequence, TypeVar

import attr
from typing_extensions import Literal

from . import error_utils, pretty_printer, type_utils
from .deserializer import Deserializer, DeserializerFactory
from .error_utils import ContextStack, Err, Ok, Result
from .types import Primitive, T, Tag

S = TypeVar("S", bound=enum.Enum)


@attr.s(auto_attribs=True, frozen=True)
class EnumValueDeserializer(Deserializer[S]):
    """Deserialize a single enum value."""

    _ds: Deserializer[S]
    _match_value: Any
    _return_value: S

    def ds(self, cs: ContextStack, value: Primitive) -> Result[S]:
        """Deserialize a single enum value."""
        v = self._ds.ds(cs, value)
        if v.is_err():
            return v
        if v.unwrap() == self._match_value:
            return Ok(self._return_value)
        return Err(cs, "value mismatch")

    def __str__(self) -> str:
        """Return a string representation of the match value."""
        return repr(self._match_value)


@attr.s(auto_attribs=True, frozen=True)
class EnumDeserializer(Deserializer[T]):
    """Deserialize a primitive into an enum value."""

    _variant_dss: Sequence[EnumValueDeserializer]

    def ds(self, cs: ContextStack, value: Primitive) -> Result[T]:
        """
        Deserialize a primitive into an enum value.

        Match the primitive against the enum's values, not its names.
        """
        variant_cs = ContextStack.new()
        v = error_utils.collect_any(
            partial(ds.ds, variant_cs.alternative(str(ds)), value)
            for ds in self._variant_dss
        )
        if v.is_ok():
            return v
        return Err(
            cs,
            f"value mismatch {pretty_printer.print_primitive(value)}: "
            f"expected one of {{{', '.join(map(str, self._variant_dss))}}}",
        )


@attr.s(auto_attribs=True, frozen=True)
class EnumDeserializerFactory(DeserializerFactory):
    """Create deserializer for an enum type."""

    def mk_ds_internal(
        self, factory: DeserializerFactory, cs: ContextStack, type_: Tag[T]
    ) -> Result[Result[Deserializer[T]]]:
        """Create deserializer for the specified enum type."""
        if not (isinstance(type_.t, type) and issubclass(type_.t, enum.Enum)):
            return Err(cs, "type mismatch")
        variants: List[T] = list(type_.t)
        # Since type_.t is an enum type, and T == type_.t, values of type T have
        # `name` and `value` attributes. However, mypy can't infer this.
        names: List[str] = [v.name for v in variants]  # type: ignore[attr-defined]
        values: List[Any] = [v.value for v in variants]  # type: ignore[attr-defined]
        return Ok(
            error_utils.collect_all(
                # pylint: disable=cell-var-from-loop
                lambda: factory.mk_ds(
                    cs.parameter_index(name, type(val)), Tag(type(val))
                )().map(lambda ds: EnumValueDeserializer(ds, val, var))
                for var, name, val in zip(variants, names, values)
            ).map(EnumDeserializer)
        )


@attr.s(auto_attribs=True, frozen=True)
class LiteralDeserializerFactory(DeserializerFactory):
    """Create deserializer for a literal type."""

    def mk_ds_internal(
        self, factory: DeserializerFactory, cs: ContextStack, type_: Tag[T]
    ) -> Result[Result[Deserializer[T]]]:
        """Create a deserializer for the specified literal type."""
        if not type_utils.get_base_of_generic_type(type_.t) == Literal:
            return Err(cs, "type mismatch")
        return Ok(
            error_utils.collect_all(
                # pylint: disable=cell-var-from-loop
                lambda: factory.mk_ds(
                    cs.type_index(_get_literal_t(v)), Tag(_get_literal_t(v))
                )().map(lambda ds: EnumValueDeserializer(ds, v, v))
                for v in type_utils.get_type_parameters(type_.t)
            ).map(EnumDeserializer)
        )


def _get_literal_t(type_: type) -> type:
    if type_utils.get_base_of_generic_type(type_) == Literal:
        return type_
    return type(type_)



"""Deserializer for a union of types."""

from functools import partial
from typing import Any, Mapping, Tuple, Union

import attr

from . import error_utils, pretty_printer, type_utils
from .deserializer import Deserializer, DeserializerFactory
from .error_utils import ContextStack, Err, Ok, Result
from .types import Primitive, T, Tag


@attr.s(auto_attribs=True, frozen=True)
class OneOfDeserializer(Deserializer[T]):
    """Deserialize with one of the supplied set of deserializers."""

    _dss: Mapping[str, Deserializer[T]]

    def ds(self, cs: ContextStack, value: Primitive) -> Result[T]:
        """Deserialize a primitive by trying several deserializers in turn."""
        return error_utils.collect_any(
            partial(ds.ds, cs.alternative(k), value) for k, ds in self._dss.items()
        ).map_err(partial(error_utils.add_err, cs.make_err("no alternative matched")))


@attr.s(auto_attribs=True, frozen=True)
class UnionDeserializerFactory(DeserializerFactory):
    """Create deserializer for a union of several types."""

    def mk_ds_internal(
        self, factory: DeserializerFactory, cs: ContextStack, type_: Tag[T]
    ) -> Result[Result[Deserializer[T]]]:
        """Create a deserializer for the specified union type."""
        if not type_utils.get_base_of_generic_type(type_.t) == Union:
            return Err(cs, "type mismatch")

        def mk_ds(i: int, t: Any) -> Result[Tuple[Any, Deserializer]]:
            return factory.mk_ds(cs.parameter_index(i, t), Tag(t))().map(
                lambda ds: (pretty_printer.print_type_name(t), ds)
            )

        return Ok(
            error_utils.collect_all(
                [
                    partial(mk_ds, i, t)
                    for i, t in enumerate(type_utils.get_type_parameters(type_.t))
                ]
            ).map(lambda dss: OneOfDeserializer(dict(dss)))
        )



"""Deserializer interface."""

import abc
from functools import partial
from typing import Callable, Generic, Mapping, TypeVar, cast

import attr
from typing_extensions import final

from . import error_utils
from .error_utils import ContextStack, Result
from .errors import DeserializationError, DeserializerFactoryError
from .types import DeserializableType, Primitive, Tag

T = TypeVar("T")


class Deserializer(Generic[T], abc.ABC):
    """Deserializer interface."""

    @final
    def __call__(self, value: Primitive) -> T:
        """Deserialize a primitive value."""
        return error_utils.unwrap(
            partial(DeserializationError, value=value),
            self.ds(ContextStack.new(), value),
        )

    @abc.abstractmethod
    def ds(self, cs: ContextStack, value: Primitive) -> Result[T]:
        """Deserialize a primitive value with the supplied context."""


class DeserializerFactory(abc.ABC):
    """Deserializer factory interface."""

    @final
    def __call__(self, type_: DeserializableType) -> Deserializer[T]:
        """Create a deserializer for the supplied type."""
        return error_utils.unwrap(
            partial(DeserializerFactoryError, type_=type_),
            cast(Result[Deserializer[T]], self.mk_ds(ContextStack.new(), Tag(type_))()),
        )

    @final
    def mk_ds(
        self, cs: ContextStack, type_: Tag[T]
    ) -> Callable[[], Result[Deserializer[T]]]:
        """Create a deserializer for the supplied type."""

        def inner() -> Result[Deserializer[T]]:
            return error_utils.join(self.mk_ds_internal(self, cs, type_))

        return inner

    @abc.abstractmethod
    def mk_ds_internal(
        self, factory: "DeserializerFactory", cs: ContextStack, type_: Tag[T]
    ) -> Result[Result[Deserializer[T]]]:
        """
        Create a deserializer for the supplied type.

        An Err return value corresponds to a mismatch between the target type
        and the factory type, with an Ok(Err) return value corresponding to
        a matching factory failing to create a deserializer.

        Recurse using the supplied factory.
        """


@attr.s(auto_attribs=True, frozen=True)
class OneOfDeserializerFactory(DeserializerFactory):
    """Deserializer factory that tries a sequence of factories in turn."""

    _factories: Mapping[str, DeserializerFactory]

    def mk_ds_internal(
        self, factory: "DeserializerFactory", cs: ContextStack, type_: Tag[T]
    ) -> Result[Result[Deserializer[T]]]:
        """Try the configured sequence of factories in turn."""
        return (
            error_utils.collect_any_it(
                partial(f.mk_ds_internal, self, cs.alternative(k), type_)
                for k, f in self._factories.items()
            )
            # pylint: disable=cell-var-from-loop
            .map(lambda dss: error_utils.collect_any(lambda: v for v in dss)).map_err(
                partial(error_utils.add_err, cs.make_err("no alternative matched"))
            )
        )



"""
Experimental features which are not guaranteed to be supported.

Breaking changes to these features do not trigger a major version bump.
"""

from .metadata import (  # noqa: F401, pylint: disable=unused-import
    DICT,
    LIST,
    VALUE,
    From,
    Keyed,
    With_,
    from_,
    keyed,
    with_,
)



"""Deserializer for a generalized homogenous sequence."""

from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    MutableSequence,
    Optional,
    Sequence,
    Set,
    Tuple,
    cast,
)

import attr

from . import error_utils, iterator_utils, type_utils
from .deserializer import Deserializer, DeserializerFactory
from .error_utils import ContextStack, Err, Ok, Result
from .types import Primitive, T, Tag


@attr.s(auto_attribs=True, frozen=True)
class SequenceDeserializer(Deserializer[T]):
    """Deserialize a list using a sequence of deserializers."""

    _untyped_ds: Deserializer[list]

    _required_dss: Sequence[Deserializer]
    _optional_dss: Sequence[Deserializer]
    _var_ds: Optional[Deserializer]

    _construct_fn: Callable[[ContextStack, Sequence[Any]], Result[T]]

    def ds(self, cs: ContextStack, value: Primitive) -> Result[T]:
        """
        Deserialize a primitive using a sequence of deserializers.

        Each list element is matched with a deserializer based on its index.
        From the start of the list:
        - Elements are first matched with the required element deserializers,
          raising an error if there are too few elements;
        - Additional elements are then matched with the optional element
          deserializers;
        - Any remaining elements are matched with the variadic element
          deserializer, if supplied. Otherwise an error is raised.
        """
        elts = self._untyped_ds.ds(cs, value)
        if elts.is_err():
            return cast(Result[T], elts)
        try:
            zipped = iterator_utils.zip_strict(
                self._required_dss, self._optional_dss, self._var_ds, elts.unwrap(),
            )
        except ValueError as e:
            return Err(cs, str(e))
        # pylint: disable=protected-access
        result = error_utils.collect_all(
            partial(ds.ds, cs.index(i), v) for i, (ds, v) in enumerate(zipped)
        )
        if result.is_err():
            return cast(Result[T], result)
        return self._construct_fn(cs, result.unwrap())


@attr.s(auto_attribs=True, frozen=True)
class HomogeneousSequenceDeserializerFactory(DeserializerFactory):
    """Create deserializer for a generalized homogenous sequence type."""

    def mk_ds_internal(
        self, factory: DeserializerFactory, cs: ContextStack, type_: Tag[T]
    ) -> Result[Result[Deserializer[T]]]:
        """Create a deserializer for the specified homogeneous sequence type."""
        sequence_fns = cast(
            Dict[T, Callable[[Sequence[Any]], T]],
            {
                FrozenSet: set,
                Iterable: list,
                Iterator: iter,
                List: list,
                MutableSequence: list,
                Sequence: list,
                Set: set,
            },
        )
        base_t = cast(T, type_utils.get_base_of_generic_type(type_.t))
        if base_t not in sequence_fns:
            return Err(cs, "type mismatch")

        [value_t] = type_utils.get_type_parameters(type_.t)
        return Ok(
            error_utils.collect_all(
                [
                    factory.mk_ds(cs, Tag(list)),
                    factory.mk_ds(cs.type_index(value_t), Tag(value_t)),
                ]
            ).map(
                lambda dss: SequenceDeserializer(
                    dss[0], [], [], dss[1], lambda _, v: Ok(sequence_fns[base_t](v))
                )
            )
        )


@attr.s(auto_attribs=True, frozen=True)
class TupleDeserializerFactory(DeserializerFactory):
    """Create deserializer for a tuple type."""

    def mk_ds_internal(
        self, factory: DeserializerFactory, cs: ContextStack, type_: Tag[T]
    ) -> Result[Result[Deserializer[T]]]:
        """Create a deserializer for the specified tuple type."""
        if not type_utils.get_base_of_generic_type(type_.t) == Tuple:
            return Err(cs, "type mismatch")
        types = type_utils.get_type_parameters(type_.t)
        if types and types[-1] == Ellipsis:
            # variadic tuple
            return Ok(
                error_utils.collect_all(
                    [factory.mk_ds(cs, Tag(list))]
                    + [factory.mk_ds(cs.type_index(t), Tag(t)) for t in types[:-1]]
                ).map(
                    lambda dss: SequenceDeserializer(
                        dss[0],
                        dss[1:-1],
                        [],
                        dss[-1],
                        lambda _, v: Ok(cast(T, tuple(v))),
                    )
                )
            )
        else:
            # fixed-length tuple
            return Ok(
                error_utils.collect_all(
                    [factory.mk_ds(cs, Tag(list))]
                    + [
                        factory.mk_ds(cs.parameter_index(i, t), Tag(t))
                        for i, t in enumerate(types)
                    ]
                ).map(
                    lambda dss: SequenceDeserializer(
                        dss[0], dss[1:], [], None, lambda _, v: Ok(cast(T, tuple(v))),
                    )
                )
            )



"""Automatically deserialize complex objects from simple Python types."""

from .errors import DeserializationError, DeserializerFactoryError  # noqa: F401
from .terramare import deserialize_into  # noqa: F401



"""Tree data structure."""

from typing import Dict, Iterator, List, MutableMapping, Tuple, TypeVar

import attr

Key = TypeVar("Key")
Value = TypeVar("Value")


@attr.s(auto_attribs=True)
class Tree(MutableMapping[Tuple[Key, ...], Value]):
    """
    Tree data structure.

    Values at nodes are optional. Children are named and ordered.
    """

    value: List[Value] = attr.ib(factory=list, init=False)
    children: Dict[Key, "Tree[Key, Value]"] = attr.ib(factory=dict, init=False)

    def __iter__(self) -> Iterator[Tuple[Key, ...]]:
        """Iterate over the tree keys."""
        if self.value:
            yield ()
        for key, child in self.children.items():
            yield from ((key,) + k for k in child)

    def __len__(self) -> int:
        """Count the number of nodes with values."""
        return sum((len(child) for child in self.children.values()), len(self.value),)

    def __getitem__(self, key: Tuple[Key, ...]) -> Value:
        """Retrieve a value from the tree by key."""
        if key:
            try:
                return self.children[key[0]][key[1:]]
            except KeyError as e:
                raise KeyError((key[0],) + e.args) from None
        if self.value:
            return self.value[0]
        raise KeyError(())

    def __setitem__(self, key: Tuple[Key, ...], value: Value) -> None:
        """Set a value by key."""
        if key:
            if key[0] not in self.children:
                self.children[key[0]] = Tree()
            self.children[key[0]][key[1:]] = value
        else:
            self.value = [value]

    def __delitem__(self, key: Tuple[Key, ...]) -> None:
        """Remove a value by key."""
        # Attempt to access the value; this will throw if it is not present.
        self[key]  # pylint: disable=pointless-statement
        if key:
            del self.children[key[0]][key[1:]]
            if not self.children[key[0]]:
                del self.children[key[0]]
        else:
            self.value = []

    def __str__(self) -> str:
        """Pretty-print a string representation of the tree."""
        return self._str(".")

    def _str(self, root: str) -> str:
        """Pretty-print the tree using the supplied name for this node."""
        if self.value:
            root += f":{_breaklines(str(self.value[0]))}"
        elements = [root]
        for i, (key, child) in enumerate(self.children.items()):
            if i == len(self.children) - 1:
                # pylint: disable=protected-access
                elements.append(_indent(child._str(str(key)), "└─ ", "   "))
            else:
                elements.append(
                    _indent(
                        child._str(str(key)),  # pylint: disable=protected-access
                        "├─ ",
                        "│  ",
                    )
                )
        return "\n".join(elements)


def _indent(s: str, head_prefix: str, tail_prefix: str) -> str:
    return "\n".join(
        (tail_prefix if lineno else head_prefix) + line
        for lineno, line in enumerate(s.splitlines())
    )


def _breaklines(s: str) -> str:
    lines = s.splitlines()
    if len(lines) <= 1:
        return " " + s
    return f"\n{_indent(s, '  ', '  ')}"



"""Common error types."""

import abc
from abc import abstractproperty

import attr

from .pretty_printer import print_primitive, print_type_name
from .tree import Tree
from .types import DeserializableType, Primitive, TerramareError


@attr.s(auto_attribs=True, frozen=True)
class _TreeException(TerramareError, abc.ABC):

    tree: Tree[str, str]

    def __str__(self) -> str:
        return f"{self.msg}\n{self.tree}"

    @abstractproperty
    def msg(self) -> str:
        """Format the exception message."""


@attr.s(auto_attribs=True, frozen=True)
class DeserializationError(_TreeException):
    """Raised when deserialization fails."""

    value: Primitive

    @property
    def msg(self) -> str:
        """Format the exception message."""
        return f"Failed to deserialize value '{print_primitive(self.value)}'"


@attr.s(auto_attribs=True, frozen=True)
class DeserializerFactoryError(_TreeException):
    """Raised when deserializer creation fails."""

    type_: DeserializableType

    @property
    def msg(self) -> str:
        """Format the exception message."""
        return f"Failed to create deserializer for type '{print_type_name(self.type_)}'"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['terramare']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.1,<20.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'result>=0.5,<0.6',
 'typing-extensions>=3.7,<4.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<0.8'],
 'test': ['coverage>=4.5,<5.0',
          'pytest>=4.0,<5.0',
          'ruamel.yaml>=0.16.5,<0.17.0',
          'toml>=0.10.0,<0.11.0']}

setup_kwargs = {
    'name': 'terramare',
    'version': '0.4.0',
    'description': 'Automatically deserialize complex objects from simple Python types',
    'long_description': '![dragon](docs/emoji.png)\n\n# terramare\n\n[![python: 3.6 | 3.7 | 3.8](https://img.shields.io/badge/python-3.6%20|%203.7%20|%203.8-blue)](https://www.python.org/)\n[![license: MIT](https://img.shields.io/badge/license-MIT-blueviolet.svg)](https://opensource.org/licenses/MIT)\n[![PyPI](https://img.shields.io/pypi/v/terramare)](https://pypi.org/project/terramare/)\n[![PyPI - Downloads](https://img.shields.io/pypi/dm/terramare)](https://pypistats.org/packages/terramare)\n[![docs: pages](https://img.shields.io/badge/docs-pages-blue)](https://tomwatson1024.gitlab.io/terramare/)\n\n[![ci status](https://gitlab.com/tomwatson1024/terramare/badges/master/pipeline.svg)](https://gitlab.com/tomwatson1024/terramare/commits/master)\n[![coverage](https://gitlab.com/tomwatson1024/terramare/badges/master/coverage.svg)](https://gitlab.com/tomwatson1024/terramare/commits/master)\n[![Checked with mypy](http://www.mypy-lang.org/static/mypy_badge.svg)](http://mypy-lang.org/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Conventional Commits](https://img.shields.io/badge/Conventional%20Commits-1.0.0-yellow.svg)](https://conventionalcommits.org)\n\nAutomatically construct complex objects from simple Python types.\n\nHighlights:\n\n- **No boilerplate:** `terramare` uses Python\'s standard type hints to determine how to construct instances of a class;\n- **Format-agnostic:** `terramare` takes simple Python types as input - pass it the output from `json.load`, `toml.load`, or `yaml.load`;\n- **Non-invasive:** `terramare` requires no modifications to your existing classes and functions beyond standard type hints;\n\nFull documentation available at <https://tomwatson1024.gitlab.io/terramare/>.\n\n## Example\n\n### Deserializing a Simple Class\n\nConsider the following simple class, defined using [`attrs`](https://github.com/python-attrs/attrs) for brevity:\n\n```python\n>>> from typing import List\n>>> import attr\n>>> import terramare\n\n>>> @attr.s(auto_attribs=True)\n... class Example:\n...     words: List[str]\n...\n...     def __str__(self):\n...         return " ".join(self.words)\n\n```\n\nDeserializing an instance of the class from a dictionary is as simple as:\n\n```python\n>>> print(terramare.deserialize_into(Example, {"words": ["hello", "world!"]}))\nhello world!\n\n```\n\n### Deserializing a More Complex Class\n\nConsider the `Person` class defined below:\n\n```python\n>>> from typing import NamedTuple, NewType, Sequence\n>>> import attr\n>>> import terramare\n\n    # `terramare` handles NamedTuples\n>>> class Location(NamedTuple):\n...     longitude: float\n...     latitude: float\n\n\n    # `terramare` handles NewType aliases\n>>> JobTitle = NewType("JobTitle", str)\n\n\n    # `terramare` handles custom classes [experimental]\n>>> @terramare.experimental.from_(terramare.experimental.DICT)\n... class Occupation:\n...     def __init__(self, title: JobTitle, field: str):\n...         self.title = title\n...         self.field = field\n...\n...     def __eq__(self, other):\n...         if isinstance(other, self.__class__):\n...             return vars(self) == vars(other)\n...         return False\n...\n...     def __repr__(self):\n...         return "Occupation(\'{0.title}\', \'{0.field}\')".format(self)\n\n\n>>> @attr.s(auto_attribs=True)\n... class Person:\n...     name: str\n...     age: int\n...     friends: Sequence[str]\n...\n...     # `terramare` handles complex member variable types\n...     location: Location\n...     occupation: Occupation\n\n```\n\nAgain, deserialization is a single function call:\n\n```python\n>>> terramare.deserialize_into(\n...     Person,\n...     {\n...         "name": "Alice",\n...         "age": 20,\n...         "friends": ["Bob", "Charlie"],\n...         "location": [51.5074, 0.1278],\n...         "occupation": {"title": "programmer", "field": "technology"}\n...     }\n... )\nPerson(name=\'Alice\', age=20, friends=[\'Bob\', \'Charlie\'], location=Location(longitude=51.5074, latitude=0.1278), occupation=Occupation(\'programmer\', \'technology\'))\n\n```\n\n## Installation\n\nInstall using [pip](https://pip.pypa.io/en/stable/):\n\n```bash\npip install terramare\n```\n\n## Alternatives\n\nCheck out:\n\n- [`pydantic`](https://pydantic-docs.helpmanual.io/) - _"Data validation and settings management using python type annotations"_. A much more mature library also using Python\'s standard type hints for deserialization that requires a little more integration with your code;\n- [`schematics`](https://schematics.readthedocs.io/en/latest/) - _"...combine types into structures, validate them, and transform the shapes of your data based on simple descriptions"_. Uses custom types instead of Python\'s standard type hints;\n- [`cerberus`](https://docs.python-cerberus.org/en/stable/) - _"...provides powerful yet simple and lightweight data validation functionality out of the box and is designed to be easily extensible, allowing for custom validation"_. Schema validation that doesn\'t change the type of the primitive value.\n',
    'author': 'Tom W',
    'author_email': '796618-tomwatson1024@users.noreply.gitlab.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/tomwatson1024/terramare',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
